;------------------------------------------------------------------------------
;-	  ATMEL Microcontroller Software Support  -   ROUSSET -
;------------------------------------------------------------------------------
;- File source				: dma_chnl.s
;- Translator				: ARM Software Development Toolkit V2.11
;-
;- Treatment				: Single-channel block DMA
;-							  The transfer is done block/block on each FIQ
;-							  One packet is a user defined 16-bytes packet
;-							  Last block is the remaining of bytes
;-
;- Imported Resources		: none
;- Exported Resources		: DmaChannelSet
;-                            DmaChannelGet
;-                            DmaChannelFiq
;-
;- 1.0 SDC 11/08/98			: Creation
;------------------------------------------------------------------------------

			AREA			AT91Lib, CODE, READONLY

;------------------------- List of Included Files -----------------------------
			INCLUDE			arm.inc

;------------------------- List of constants and types ------------------------
;- DMA Channel descriptor
INC_DEST				EQU		0x1
INC_SRC					EQU		0x2

PACKET_SIZE				EQU		4*4

;------------------------- List of Imported resources -------------------------
;- None

;------------------------- List of Internal resources -------------------------
;- None

;------------------------- List of External resources -------------------------

;------------------------------------------------------------------------------
;- Function				: DmaChannelSet
;- Treatments			: Set and enable the channel DMA
;-                        (This function must be called in PRIVILEDGE mode !)
;- Input Parameters		: r0 = dma_desc : pointer to DMA descriptor
;-                        r1 = buf : temporary buffer (7*4 bytes needed)
;- Output Parameters	: none
;- Registers lost		: r2 r3
;- Called Functions		: none
;- Called Macros		: none
;------------------------------------------------------------------------------
			EXPORT		DmaChannelSet
;- Begin
DmaChannelSet

;- | Switch to FIQ mode and save entry mode
			mrs			r2, CPSR
			mov			r3, #ARM_MODE_FIQ:OR:I_BIT:OR:F_BIT
			msr			CPSR, r3

;- | r13_fiq <- temporary buffer pointer
			mov			r13, r1

;- | Load banked registers from the DMA descriptor
			ldmia			r0, {r8 - r12}

;- | Set default number of packets to transfer (1) if not defined by user
			cmp			r11, #0
			moveq		r11, #1

;- | Enable DMA (Clear bit F) if number of bytes to transfer > 0
			cmp			r10, #0
			bicne		r2, r2, #F_BIT

;- | Restore entry mode
			msr			CPSR, r2
;- End
			mov			pc,lr

;------------------------------------------------------------------------------
;- Function				: DmaChannelGet
;- Treatments			: get the count of remaining bytes to be transferred
;-                        (This function must be called in PRIVILEDGE mode !)
;- Input Parameters		: none
;- Output Parameters	: r0 = count : Number of bytes remaining
;- Registers lost		: r2, r3
;- Called Functions		: none
;- Called Macros		: none
;------------------------------------------------------------------------------
			EXPORT		DmaChannelGet
;- Begin
DmaChannelGet

;- | Switch to FIQ mode and save entry mode
			mrs			r2, CPSR
			mov			r3, #ARM_MODE_FIQ:OR:I_BIT
			msr			CPSR, r3

;- | Return number of bytes remaining to be transferred
			movs		r0, r10

;- | Restore entry mode
			msr			CPSR, r2

;- End
			mov			pc,lr

;------------------------------------------------------------------------------
;- Function				: DmaChannelFiq
;- Treatments			: FIQ transfert interruption
;-
;- Input Parameters		: r8  = src : Source Buffer pointer
;-						  r9  = dest : Desctination Buffer pointer
;-						  r10 = nb_bytes : Total number of bytes to transfer
;-						  r11 = nb_packets : Number of packets to transfer at each FIQ
;-						  r12 = inc_mask : Pointer increment flags
;-						  r13 = Temporary buffer
;- Output Parameters	: none
;- Registers lost		: none
;- Called Functions		: none
;- Called Macros		: none
;------------------------------------------------------------------------------
			EXPORT		DmaChannelFiq
;- Begin
DmaChannelFiq

;- | Save used registers in the temporary buffer
			stm  		r13, {r0 - r3, r6 - r7, r11}
;- | Save source and destination
			mov			r6, r8
			mov			r7, r9

;- | While (nb_packets > 0) and (nb_bytes >= PACKET_SIZE)
Fiq_copy_packet
;- | | Update nb_bytes by sustracting PACKET_SIZE from it 
			subs		r10, r10, #PACKET_SIZE
			blt			Fiq_copy_bytes
;- | | *dest <- *src (multiple load and store 4 words with post increment)
			ldmia		r8!, {r0-r3}
			stmia		r9!, {r0-r3}
;- | | Decrement nb_packets
			subs		r11, r11, #1
;- | EndWhile
			bne			Fiq_copy_packet

;- | If (nb_packets == 0) (i.e. All blocks tranfered)
;- | Then
			b			Fiq_update
Fiq_copy_bytes
			add			r10, r10, #PACKET_SIZE
			beq			Fiq_disable
;- | | === Copy last bytes to copy (< 16) ===
;- | | *dest <- *src
;- | | If needed, load and store multiple 2 words with post increment
			ands		r0, r10, #0x08
			ldmneia		r8!, {r0-r1}
			stmneia		r9!, {r0-r1}

;- | | If needed, load and store 1 word with post increment
			ands		r0, r10, #0x04
			ldrne		r0, [r8], #4
			strne		r0, [r9], #4

;- | | If needed, load and store 1 half-word with post increment
			ands		r0, r10, #0x02
			ldrneh		r0, [r8], #2
			strneh		r0, [r9], #2

;- | | If needed, load and store 1 byte with post increment
			ands		r0, r10, #0x01
			ldrneb		r0, [r8], #1
			strneb		r0, [r9], #1
Fiq_disable

;- | | Clear nb_bytes
			mov			r10, #0
;- | | Disable DMA (SPSR.F <- 0)
			mrs			r0, SPSR
			orr			r0, r0, #F_BIT
			msr			SPSR, r0
;- | Endif
Fiq_update

;- | Restore src (if not INC_SRC)
			ands		r0, r12, #INC_SRC
			moveq		r8, r6
;- | Restore destination (if not INC_DEST)
			ands		r0, r12, #INC_DEST
			moveq		r9, r7

;- | Restore used registers
			ldm			r13, {r0 - r3, r6 - r7, r11}
;- End
			subs		pc, lr, #4

			END
